// $Id: z80_mem.c,v 1.6 2011-04-03 17:23:52 drmiller Exp $

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "z80_mem.h"

/// \brief Locate physical memory for virtual address
///
/// \param[in] mem	System memory object
/// \param[in] adr	Virt address
///
static uint8_t *find_map_adr(z80_mem_t *mem, uint16_t adr) {
	mem_map_t *map;
	map = mem->mmu[adr >> 8];
	if (!map) {
		mem->segfault(mem->sys, adr);
	}
	return &map->mem[adr - map->vadr];
}

/// \brief Install a bank of memory into virt addr space
///
/// \param[in] mem	System memory object
/// \param[in] map	Bank of memory to enable
///
static void mapin(z80_mem_t *mem, mem_map_t *map) {
	uint32_t a;
	for (a = map->vadr; a < map->vadr + map->len; a += 0x0100) {
		mem->mmu[a >> 8] = map;
	}
}

/// \brief Setup a bank of memory
///
/// The caller must have initialized mem before first call.
/// This typically entails zeroing plus setting segfault and sys.
///
/// \param[in] mem	System memory object
/// \param[in] padr	Phys Addr of new memory (page aligned)
/// \param[in] vadr	Virt Addr of memory (page aligned)
/// \param[in] len	Length of bank (pages, 1-256)
/// \param[in] src	(optional) Filename of memory contents
/// \param[in] sadr	(optional) Load address within bank
///
void mem_init(z80_mem_t *mem, uint32_t padr, uint16_t vadr, uint32_t len,
						char *src, uint16_t sadr) {
	// assert((padr & 0x000000ff) == 0);
	// assert((vadr & 0x000000ff) == 0);
	// assert((len & 0x000000ff) == 0);
	// assert((len & 0xffffff00) != 0);
	// assert((len & 0xffffff00) <= 256);
	// assert: no map exists in padr range...
	mem_map_t *map;
	map = malloc(sizeof(*map));
	memset(map, 0, sizeof(*map));
	map->padr = padr;
	map->vadr = vadr;
	map->len = len;
	map->mem = malloc(map->len);
	memset(map->mem, 0, map->len);
	if (src) {
		map->src = src;
		int fd;
		fd = open(src, O_RDONLY);
		if (fd < 0) {
			perror(src);
			exit(1);
		}
		(void)read(fd, map->mem + sadr, map->len - sadr);
		// assert(n <= map->len - sadr);
		close(fd);
	}
	map->next = mem->maps;
	mem->maps = map;
	mapin(mem, map);
}

/// \brief Map in a bank of physical memory
///
/// Enables all maps within the physical address range.
/// Overlaps are considered a mistake, but silently ignored.
/// The last overlapping bank encountered will prevail.
///
/// \param[in] mem	System memory object
/// \param[in] padr	Phys Addr of bank(s) to map in
/// \param[in] len	Phys Addr of bank(s) to map in
///
void mem_mmu_map(z80_mem_t *mem, uint32_t padr, uint32_t len) {
	mem_map_t *map;
	for (map = mem->maps; map; map = map->next) {
		if (map->padr >= padr && map->padr + map->len <= padr + len) {
			mapin(mem, map);
		}
	}
}

uint8_t *mem_get_addr(z80_mem_t *mem, uint16_t adr) {
	uint8_t *m = find_map_adr(mem, adr);
	return m;
}
uint8_t mem_get_byte(z80_mem_t *mem, uint16_t adr) {
	uint8_t *m = find_map_adr(mem, adr);
	return *m;
}
uint16_t mem_get_word(z80_mem_t *mem, uint16_t adr) {
	uint16_t *m = (uint16_t *)find_map_adr(mem, adr);
	return *m;
}
