// $Id: z80_sys.c,v 1.15 2011-04-08 23:11:31 drmiller Exp $

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>

#include "z80_sys.h"
#include "z80_exec.h"

#ifdef USE_BDOS
#include "bdos.h"
#endif // USE_BDOS

extern int diz80(char *buf, uint8_t *t);

static int intr(z80_sys_t *sys, int sig) {
	if (sig == SIGINT && sys->run) {
		sys->run = 0;
		return 0;
	}
	return 1;
}

static void dump(z80_sys_t *sys) {
	char buf[1024];
	char *s = buf;
	fprintf(stderr, "at cycle %lld\n", sys->cpu.cycles);

	// todo: call z80_sim_cmd.c:_dump()...
	uint8_t *pc = mem_get_addr(&sys->mem, sys->cpu.pc);

	diz80(buf, pc);
	fprintf(stderr, "PC = %04x [ %s ]\n", sys->cpu.pc, buf);
	uint16_t *sp = (uint16_t *)mem_get_addr(&sys->mem, sys->cpu.sp);
	fprintf(stderr, "SP = %04x [ %04x %04x %04x %04x ]\n",
			sys->cpu.sp, sp[0], sp[1], sp[2], sp[3]);
	uint16_t bc = sys->cpu.gp_regs.gp[sys->cpu.gp_regs.gp_f & 1].xx[rp_BC];
	uint16_t de = sys->cpu.gp_regs.gp[sys->cpu.gp_regs.gp_f & 1].xx[rp_DE];
	uint16_t hl = sys->cpu.gp_regs.gp[sys->cpu.gp_regs.gp_f & 1].xx[rp_HL];
	fprintf(stderr, "BC = %04x DE = %04x HL = %04x\n", bc, de, hl);
	// more...
}

static void iofault(void *v, uint8_t port) {
	z80_sys_t *sys = (z80_sys_t *)v;
	fprintf(stderr, "I/O Fault at port %02x\n", port);
	dump(sys);
	exit(1);
}

static void segfault(void *v, uint16_t adr) {
	z80_sys_t *sys = (z80_sys_t *)v;
	fprintf(stderr, "Seg Fault at address %04x\n", adr);
	dump(sys);
	exit(1);
}

static void sysfault(z80_sys_t *sys, const char *str) {
	fprintf(stderr, "%s\n", str);
	dump(sys);
	exit(1);
}

void sys_init(z80_sys_t *sys) {
	memset(sys, 0, sizeof(*sys));
	sys->fault = sysfault;
	//cpu_init(&sys->cpu);
	sys->io.iofault = iofault;
	sys->io.sys = sys;
	sys->mem.segfault = segfault;
	sys->mem.sys = sys;
	// now install all memory banks and maps...
	mem_init(&sys->mem, 0x00000000, 0x0000, 64 * 1024, NULL, 0);
#ifdef USE_BDOS
	sys->ccp = NULL;
	bdos_init(sys);
#endif // USE_BDOS
	sys->intr = intr;
	sys->trace = 0;
	sys->trc_fp = stderr;

	// now install all devices and peripherals...
}

static void sys_interact(z80_sys_t *sys) {
	extern void sys_command(z80_sys_t *sys);

	printf("break at %04x\n", sys->cpu.pc);
	do {
		sys_command(sys);
	} while (!sys->run);
}

#ifdef USE_BDOS
void sys_setccp(z80_sys_t *sys, char *exe, uint16_t adr, uint16_t entry) {
	int fd;
	struct stat stb;
	fd = open(exe, O_RDONLY);
	if (fd < 0) {
		perror(exe);
		exit(1);
	}
	fstat(fd, &stb);
	if (stb.st_size >= 64*1024 ||
			(off_t)adr + stb.st_size > 64*1024) {
		fprintf(stderr, "CCP too big (%ld @ %04x)\n", stb.st_size, adr);
		exit(1);
	}
	sys->ccp_adr = adr;
	sys->ccp_ent = entry;
	sys->ccp_len = stb.st_size & 0x0000ffff;
	sys->ccp = malloc(sys->ccp_len);
	if (!sys->ccp) {
		perror(exe);
		exit(1);
	}

	int rc = read(fd, sys->ccp, sys->ccp_len);
	if (rc < 0) {
		perror(exe);
		exit(1);
	}
	close(fd);
}
#endif // USE_BDOS

void sys_loadpgm(z80_sys_t *sys, char *exe, uint16_t adr, uint16_t entry) {
	int fd;
	uint32_t max = 64 * 1024;
	fd = open(exe, O_RDONLY);
	if (fd < 0) {
		perror(exe);
		exit(1);
	}
	// this doesn't work for banked memory!
	uint8_t *m = mem_get_addr(&sys->mem, adr);
	(void)read(fd, m, max - adr);
	close(fd);
}

static void run_some(z80_sys_t *sys, uint16_t entry) {
	sys->cpu.pc = entry;
	do {
		while (!sys->run) {
			sys_interact(sys);
		}
#ifdef USE_BDOS
		if (bdos_trap(sys) == 0)
#endif // USE_BDOS
		{
			sys_exec(sys); // single-step
		}
	} while (sys->cpu.pc != 0x0000);
}


void sys_go(z80_sys_t *sys, uint16_t entry) {
	run_some(sys, entry);
	printf("BDOS exit at cycle %lld\n", sys->cpu.cycles);
#ifdef TRACE
	if (sys->trace) {
		dump(sys);
	}
#endif // TRACE
}

#ifdef USE_BDOS
void sys_runccp(z80_sys_t *sys) {
	do {
		while (!sys->run) {
			sys_interact(sys);
		}
		// this doesn't work for banked memory!
		uint8_t *m = mem_get_addr(&sys->mem, sys->ccp_adr);
		memcpy(m, sys->ccp, sys->ccp_len);
		*(get_gp(sys, gp_C)) = 0;	// user/drive
		run_some(sys, sys->ccp_ent);
	} while (1); // todo: find other exit criteria
	printf("BDOS exit at cycle %lld\n", sys->cpu.cycles);
#ifdef TRACE
	if (sys->trace) {
		dump(sys);
	}
#endif // TRACE
}
#endif // USE_BDOS
